/*
 * Copyright (c) 2003 Andrew Stikheev. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * $Id: scand.c,v 1.14 2003/06/01 18:34:03 sand Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <syslog.h>
#include <sys/sysctl.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <signal.h>

#include <fcntl.h>
#include <time.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netdb.h>

#include <math.h>
#include <err.h>
#include <errno.h>
#include <string.h>

#include "statistics.h"
#define PIDFILE "/var/run/scand.pid"

//--------------------------------------------------
/*
 * Function prototypes.
 */
static void		Warn (const char*);
static void		DaemonMode ();
static void		InitiateShutdown (int);
static void		Shutdown (int);
static void		ParseArgs (int argc, char **argv);
static void		Usage (char *);
static void		KillDaemon ();
static void		Quit ( const char* );

/*
 * Globals.
 */
clusterItem		*hashTable;

static	int		verbose		= 0; // disabled
static	int		background  	= 0; // not a daemon
static	int		running     	= 1; 
static	u_short		divertPort   	= 123;
static  char		*mmapedFile 	= NULL;
int			enableBlock	= 0;
int			fdLog 		= -1;

char                    packetBuf[IP_MAXPACKET];

////////////////////////////////////////////////////////////////////////////////
int main (int argc, char** argv)
{
	int                     packetLen,write;
	struct sockaddr_in      packetAddr;

	int			divert;
	int      		addrSize; 
	struct sockaddr_in	addr;
	long    		writePacketsFailure = 0;
	long    		blockedPackets = 0;

/*
 * Parse options.
 */
	ParseArgs(argc, argv);

/*
 * Open syslog channel.
 */
	openlog ("scand", LOG_CONS | LOG_PID | (verbose ? LOG_PERROR : 0),
		 LOG_DAEMON);

//-----------------------------------------------------------------------
	divert = socket (PF_INET, SOCK_RAW, IPPROTO_DIVERT);
	if (divert == -1)
		Quit ("Unable to create divert socket.");
/*
 * Bind socket.
 */

	addr.sin_family		= AF_INET;
	addr.sin_addr.s_addr	= INADDR_ANY;
	addr.sin_port		= htons(divertPort);

	if ( bind (divert, (struct sockaddr*) &addr, sizeof addr) == -1 )
		Quit ("Unable to bind divert socket.");


	if ( !verbose )	DaemonMode ();
/*
 * Catch signals to manage shutdown 
 */
	siginterrupt(SIGTERM, 1);
	siginterrupt(SIGHUP, 1);
	signal (SIGTERM, InitiateShutdown);

	if (verbose) fprintf(stderr,"start pipe\n");
//------------------------------------------------------------------------------

	while (running) {

		/*
 		 * Get packet from socket.
		 */
		addrSize = sizeof packetAddr;
        	packetLen = recvfrom (	divert,
					packetBuf,
                              		sizeof packetBuf,
                              		0,
                              		(struct sockaddr*) &packetAddr,
                              		&addrSize );

        	if ( packetLen == -1 ) continue;

		/*
		 * Handle packet.
		 */
        	if ( AddPacketToStat( packetBuf ) && enableBlock ) {
		
			blockedPackets++;
			continue;
		}

		/*
 		 * Put packet back for processing.
 		 */
		write =	sendto (	divert, 
					packetBuf,
       					packetLen, 
					0,
					(struct sockaddr*) &packetAddr,
					addrSize );

		if ( write != packetLen )  writePacketsFailure++;
	}
//------------------------------------------------------------------------------
	if (background)
		unlink (PIDFILE);

	if ( fdLog >= 0  ) close(fdLog);

	if ( mmapedFile ) {

		unlink(mmapedFile);
	}

	return 0;
}

////////////////////////////////////////////////////////////////////////////////
static void DaemonMode ()
{
	FILE*	pidFile;

 	daemon(0,0);
	background = 1;

	pidFile = fopen (PIDFILE, "w");
	if (pidFile) {

		fprintf (pidFile, "%d\n", getpid ());
		fclose (pidFile);
	}
}

////////////////////////////////////////////////////////////////////////////////
void Quit (const char* msg)
{
	Warn (msg);
	exit (1);
}

////////////////////////////////////////////////////////////////////////////////
void Warn (const char* msg)
{
	if (background)
		syslog (LOG_ALERT, "%s (%m)", msg);
	else
		warn (msg);
}
////////////////////////////////////////////////////////////////////////////////
static void InitiateShutdown (int sig)
{
/*
 * Start timer to allow kernel gracefully
 * shutdown existing connections when system
 * is shut down.
 */
	siginterrupt(SIGALRM, 1);
	signal (SIGALRM, Shutdown);
	alarm (10);
}

////////////////////////////////////////////////////////////////////////////////
static void Shutdown (int sig)
{
	running = 0;
}

////////////////////////////////////////////////////////////////////////////////
static void KillDaemon()
{
	FILE *pidFile;
	long p;

	pidFile = fopen( PIDFILE, "r" );
	if ( !pidFile )
		Quit("fopen"); 
	
	if ( !fscanf( pidFile , "%ld", &p ) )
		Quit("fscanf");

	if ( kill( (pid_t) p , SIGTERM ) )
		Quit("kill");         

	printf("Signal SIGTERM to process %ld was send\n", p );
	exit(1);
}
////////////////////////////////////////////////////////////////////////////////
static void Usage(char *error)
{
 	if (*error) fprintf(stderr,"\n%s\n",error);
 	fprintf(stderr,"Usage :\n");
	fprintf(stderr,"    -v              -  verbose mode\n");
	fprintf(stderr,"    -p port         -  divert port\n");
	fprintf(stderr,"    -l stdout       -  put statistic to stdout\n");
	fprintf(stderr,"    -l logfile      -  put statistic to logfile\n");
	fprintf(stderr,"    -m mmapfile     -  file to mmaping\n");
	fprintf(stderr,"    -q              -  send SIGTERM to last daemon\n");
	fprintf(stderr,"    -b              -  enable blocking mode\n");
	fprintf(stderr,"    -h              -  current help\n");
	exit(1);
} 
////////////////////////////////////////////////////////////////////////////////
static void ParseArgs(int argc, char ** argv)
{
	char  *end, ch;
	int fdMmapedFile = -1;

	while ((ch = getopt(argc, argv,"hvqbl:p:m:")) != -1)
		switch (ch) {
		case 'v':
				verbose = 1;
				break;
		case 'p':
				divertPort = strtol(optarg,&end,10);
				if ( end == optarg) Usage("invalid port");
				break;   	
		case 'l':	
				if ( !strcmp(optarg,"stdout") ) {

				 	fdLog = fileno (stdout);
					if ( fcntl(fdLog,F_SETFL,O_NONBLOCK)
					     == -1 ) {
				
						Quit("fcntl");
					}
				} else if ((fdLog=open( optarg,
						  	O_NONBLOCK | O_CREAT |
						  	O_WRONLY,
							S_IRUSR    | S_IWUSR |
						  	S_IRGRP    | S_IROTH
					   )) < 0 )  Quit("open");	

				if ( lseek(fdLog, 0 ,SEEK_END) == -1)
					Quit ("lseek");
				break;
		case 'm':	
				fdMmapedFile = open(	optarg,
							O_RDWR     | O_CREAT |
							O_TRUNC,
							S_IRUSR    | S_IWUSR |
							S_IRGRP    | S_IROTH ); 
				if ( fdMmapedFile < 0 ) 
					Quit ("Unable to open file to mmap.");

				if ( ftruncate(fdMmapedFile, NUM_CLUSTERS 
						* sizeof (clusterItem))   )
					Quit("ftruncate");

				hashTable = (clusterItem *) mmap( 0, 
					NUM_CLUSTERS * sizeof (clusterItem),
					PROT_READ | PROT_WRITE,
					MAP_NOSYNC | MAP_SHARED,
					fdMmapedFile,
					0);
				if ( errno ) 
					Quit("mmap");

				if ( close(fdMmapedFile) == -1 ) 
					Quit("close"); 

				mmapedFile = (char *) malloc( (strlen(optarg)+1)
							      * sizeof (char));
				if ( !mmapedFile ) 
					Quit("malloc");

				strcpy(mmapedFile,optarg); 

				break;

		case 'b':	enableBlock = 1;
				break;
		case 'q':
				KillDaemon();			
		case '?': 
		case 'h':
		default :
				Usage("");
		}
	
if (fdMmapedFile == -1 ) { 

	hashTable = (clusterItem *)malloc(NUM_CLUSTERS * sizeof(clusterItem ));
	if (!hashTable) Quit("Not enough memory.");
}

if (verbose) fprintf( stderr,"verbose = %d   port = %d put = %d\n", verbose, divertPort, fdLog ); 
}
////////////////////////////////////////////////////////////////////////////////
